#include <stdio.h>
#include <stdlib.h>
#include <cyg/io/devtab.h>

#include "gfx.h"
#include "kbd.h"
#include "libjpeg/jpeglib.h"

#include "byte2sec.h"

extern int imagesize;

/* Handle for the floppy drive IO system. */
cyg_io_handle_t fh;

struct JpegInfo {
    unsigned char *addr;
    int size;
};

#define READLEN 2048

/* The following are necessary to fool eCos into linking correctly.
 * Not having them does not seem to cause problems. */
void*   __dso_handle = (void*) &__dso_handle;

void
__cxa_atexit(void (*arg1)(void*), void* arg2, void* arg3)
{
}

/*
 *-----------------------------------------------------------------------------
 *
 * floppy_raw_read --
 *
 * 	Reads *len bytes into buf at offset (in bytes) pos.
 *
 * Results:
 *	Result of read operation.
 *
 * Side Effects:
 *
 *	The two bytes/sectors functions transform bytes into sectors,
 *	and then back, because the block IO system uses sector
 *	numbers.  The two functions also take care to allocate a chunk
 *	of memory that is big enough to handle the whole read.
 *
 *-----------------------------------------------------------------------------
 */

int floppy_raw_read(char *buf, unsigned int *len, unsigned int pos)
{
    char *newbuf = NULL;
    unsigned int newpos = 0;
    unsigned int newlen = 0;
    int retcode = 0;

    bytes2sectors(*len, pos, &newbuf, &newlen, &newpos);
    retcode = cyg_io_bread(fh, newbuf, &newlen, newpos);
    sector2bytes(newbuf, newpos, buf, len, pos);
    return retcode;
}


/*
 *-----------------------------------------------------------------------------
 *
 * read_archive_header --
 *
 * 	Reads the jpeg archive header, which is structured like so:
 *
 *	int - number of jpegs in the archive.
 *      int - 1st jpeg start offset
 *      int - 2nd jpeg start offset
 *      int - nth jpeg start offset
 *      int - nth jpeg end offset
 *
 *	All int's are little endian.
 *
 * Results:
 *	Returns the numbe of images in the archive.
 *
 * Side Effects:
 *
 *	Fills the jarch array with a list of offset positions.
 *
 *-----------------------------------------------------------------------------
 */

int read_archive_header(char *buf, unsigned int **jarch)
{
    int i = 0;
    int imgcnt = 0;

    memcpy(&imgcnt, buf, sizeof(int));
    buf += sizeof(int);
    diag_printf("%d images in the archive\n", imgcnt);
    *jarch = malloc((imgcnt + 1) * sizeof(int));
    if (*jarch == NULL) {
	diag_printf("malloc failed in read_archive_header\n");
	exit(1);
    }

    for (i = 0; i <= imgcnt; i++) {
	memcpy(*jarch + i, buf, sizeof(int));
	buf += sizeof(int);
    }
    return imgcnt;
}

/* This information is given to us by the eCos boot code. */

typedef struct screen_info {
    unsigned char *base_buf;	/* Base address of video memory. */
    int width;			/* Screen width. */
    int height;			/* Screen height. */
    int bpp;			/* Bytes per pixel. */
    int scanline;		/* Bytes per 'scanline' (one
				 * horizontal line of the screen). */
} screen_info;

/*
 *-----------------------------------------------------------------------------
 *
 * get_screen_info --
 *
 * 	Get information about the display and put it in storage that
 * 	isn't about to be wiped out.
 *
 * Results:
 *	None.
 *
 *-----------------------------------------------------------------------------
 */

get_screen_info(screen_info *si)
{
    /* This is screen information. */
    si->base_buf = (unsigned char*) current_mode->physical_base_ptr;
    si->scanline = info_block->version >= 0x0300
	? current_mode->linear_bytes_per_scanline :
	current_mode->bytes_per_scanline;
    si->bpp = current_mode->bits_per_pixel / 8;
    si->width = current_mode->x_resolution;
    si->height = current_mode->y_resolution;
}

/*
 *-----------------------------------------------------------------------------
 *
 * clear_screen --
 *
 * 	Clears the screen and fills it with a shade of gray.
 *
 *-----------------------------------------------------------------------------
 */

clear_screen(screen_info *si, unsigned color)
{
    memset(si->base_buf, color, si->scanline * si->height - 1);
}


/*
 *-----------------------------------------------------------------------------
 *
 * get_jpeg_image --
 *
 * 	This does the real work of uncompressing the jpeg and copying
 * 	it to the display memory area.
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *
 *	Uncompresses the jpeg, copy it to display memory.
 *
 *-----------------------------------------------------------------------------
 */

get_jpeg_image(struct screen_info *si,
	       unsigned char *jpegdata,
	       int jpegsize)
{
    struct jpeg_decompress_struct cinfo;
    JSAMPARRAY buffer;		/* Output row buffer */
    int row_stride;		/* physical row width in output buffer */
    struct jpeg_error_mgr jerr;
    unsigned int imgcnt = 0;
    int x, y;
    int ystart;
    int xstart;
    int imgw, imgh;
    int bpp;

    cinfo.err = jpeg_std_error(&jerr);
    jpeg_create_decompress(&cinfo);

    /* In order to use memory locations as a source, it was necessary
     * to create a custom driver for the jpeg library.  This is the
     * call to initialize the special read methods I created. */
    jpeg_mem_src(&cinfo, jpegdata, jpegsize);

    (void) jpeg_read_header(&cinfo, TRUE);

    /* After this, we should have some information about the JPEG. */
    (void) jpeg_start_decompress(&cinfo); 

    row_stride = cinfo.output_width * cinfo.output_components;
    buffer = (*cinfo.mem->alloc_sarray)
	((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);

    if (buffer == NULL)
	return;

    imgw = cinfo.output_width;
    imgh = cinfo.output_height;

    /* These center the image on the screen. */
    if (si->height >= imgh) {
	ystart = si->height/2 - imgh/2;
    } else {
	ystart = 0;
    }
    if (si->width >= imgw) {
	xstart = si->width/2 - imgw/2;
    } else {
	xstart = 0;
    }

    bpp = cinfo.output_components;

    y = 0;
    while (cinfo.output_scanline < cinfo.output_height) {
	/* jpeg_read_scanlines expects an array of pointers to scanlines.
	 * Here the array is only one element long, but you could ask for
	 * more than one scanline at a time if that's more convenient.
	 */
	unsigned char *imgbuf;
	unsigned char *line_buf = NULL;

	jpeg_read_scanlines(&cinfo, buffer, 1);

	if (imgh > si->height) {
	    if (cinfo.output_scanline < (imgh - si->height) / 2) {
		continue;
	    }
	}
	line_buf = si->base_buf + (si->scanline * y) + (si->scanline * ystart);

	/* If the image is bigger than the screen, clip it. */
	if (imgw > si->width) {
	    imgcnt = (imgw/2 - si->width/2) * bpp;
	    imgcnt /= 2;
	} else {
	    imgcnt = 0;
	}
	imgbuf = buffer[0];
	imgbuf += imgcnt;

	for (x = xstart; x < si->width; x++)
	{
	    unsigned char *buf = line_buf + (si->bpp * x);

 	    if (x < xstart + imgw) {
		*buf = imgbuf[imgcnt+2];
		*(buf + 1) = imgbuf[imgcnt + 1];
		*(buf + 2) = imgbuf[imgcnt];
		*(buf + 3) = 0;
		imgcnt += 3;
	    }
	}
 	if (y >= imgh + ystart)
	    break;

	/* If we have run out of screen, we still have to finish
	 * hoovering up the data so that libjpeg is happy. */
	if (y >= si->height) {
	    while (cinfo.output_scanline < cinfo.output_height) {
		jpeg_read_scanlines(&cinfo, buffer, 1);
	    }
	    break;
	}
	y++;
    }

    (void) jpeg_finish_decompress(&cinfo);
    jpeg_destroy_decompress(&cinfo);
}


int
main(int argc, char** argv)
{
    /* eCos puts the information we need here.   */
    info_block = (struct VBESVGAInfoBlock*) 0x000A0000;
    current_mode = (struct VBEModeInfoBlock*) 0x000A0200;
    screen_info si;

    int i = 0;
    int res;
    char buf[READLEN];
    int len = READLEN;
    unsigned int *jarch;
    unsigned int jpegsize = 0;
    unsigned int offset;
    int imgcnt = 0;
    int delay = 1000;
    struct JpegInfo *jpegs;

    get_screen_info(&si);
    clear_screen(&si, 128);

    if (cyg_io_lookup("/dev/fd", &fh) != ENOERR) {
	diag_printf("Cannot find /dev/fd, aborting\n");
	exit(1);
    }
    if ((res = floppy_raw_read(buf, &len, imagesize)) != ENOERR) {
	diag_printf("Error reading from /dev/fd - %d\n", res);
	exit(1);
    }
    if (len != READLEN) {
	diag_printf("EOF\n");
	exit(1);
    }

    imgcnt = read_archive_header(buf, &jarch);
    jpegs = malloc(sizeof(struct JpegInfo) * imgcnt);
    if (jpegs == NULL) {
	diag_printf("Failed to malloc space for %d JpegInfo structs\n", imgcnt);
	exit(1);
    }


    while (i < imgcnt) {
	unsigned char *jpegdata;
	offset = jarch[i];
	jpegsize = jarch[i + 1] - jarch[i];
	jpegdata = malloc(jpegsize);
	if (jpegdata == NULL) {
	    diag_printf("Malloc failed! %d\n", jpegsize);
	    exit(1);
	}

	len = jpegsize;

 	if ((res = floppy_raw_read(jpegdata, &len, imagesize + offset)) != ENOERR) {
	    diag_printf("Error reading from /dev/fd - %d\n", res);
	    exit(1);
	}
 	if (len < jpegsize) {
	    diag_printf("EOF - %d\n", len);
	    exit(1);
	}

	jpegs[i].addr = jpegdata;
	jpegs[i].size = jpegsize;
	clear_screen(&si, 128);
	get_jpeg_image(&si, jpegdata, jpegsize);

	i ++;
    }

    setup_kbd();

    while (1) {
	int flag = 0;
	for (i = 0; i < imgcnt; i++) {
	    flag = 0;
	    clear_screen(&si, 128);
	    get_jpeg_image(&si, jpegs[i].addr, jpegs[i].size);
	    flag = cyg_flag_timed_wait(
		&kbd_flag,
		SPACE_FLAG | FASTER_FLAG | SLOWER_FLAG,
		CYG_FLAG_WAITMODE_OR|CYG_FLAG_WAITMODE_CLR,
		cyg_current_time() + delay);

	    if (flag & FASTER_FLAG) {
		delay -= 100;
		if (delay < 0)
		    delay = 0;
	    } else if (flag & SLOWER_FLAG) {
		delay += 100;
	    }

	}
    }
}
